#pragma once

#include <cppcore/misc/string_builder.h>

#include <cppmp/core/types.h>
#include <cppmp/core/checker.h>
#include <cppmp/core/modifier.h>
#include <cppmp/core/conditionals.h>
#include <cppmp/misc/generic_predicate.h>

#include <cpprtti/types/class_type.h>

#include "variant.h"
#include "exception.h"
#include "../registry.h"

namespace cpprtti
{

    namespace __impl
    {

        /* variant_impl */

        variant_impl::variant_impl(
            const cpprtti::type& p_type)
            : type(p_type)
            { }

        void variant_impl::assign(const variant& v)
        {
            if (is_const)
            {
                throw exception(
                    error_code::variant_is_const,
                    "The value stored in the variant is const!");
            }

            if (v.empty())
            {
                throw exception(
                    error_code::variant_is_empty,
                    "The passed variant does not store any value!");
            }

            if (type.id() != v.type().id())
            {
                auto converter = v.type().convert_to(type.id());
                if (!converter)
                {
                    throw exception(
                        error_code::variant_convert_to,
                        cppcore::string_builder()
                            << "No known conversion from type '"
                            << v.type().name()
                            << "' to type '"
                            << type.name()
                            << "'!");
                }
                this->assign_impl((*converter)(v));
            }
            else
            {
                this->assign_impl(v);
            }
        }

        /* variant_builder - default */

        template<typename X, typename>
        struct variant_builder
        {
            using is_default = cppmp::true_t;

            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for make_variant(...)!"); }
        };

        /* variant_builder - not lvalue reference */

        template<typename T_registry, typename T>
        struct variant_builder<
            cppmp::list<T_registry, T>,
            cppmp::enable_if_t<
                    cppmp::is_same_v<cppmp::decay_t<T_registry>, registry>
                && !cppmp::is_lvalue_reference_v<T>
            >
        >
        {
            struct impl
                : public variant_impl
            {
                T storage;

                template<typename X>
                inline impl(const cpprtti::type& p_type, X&& x)
                    : variant_impl  (p_type)
                    , storage       (std::forward<X>(x))
                {
                    data            = reinterpret_cast<void*>(const_cast<cppmp::decay_t<T>*>(&storage));
                    is_const        = cppmp::is_const_v<cppmp::remove_reference_t<T>>;
                    is_reference    = false;
                }

                void assign_impl(const variant& v) override
                {
                    *static_cast<cppmp::decay_t<T>*>(data) = v.get<const T&>();
                };
            };

            template<typename X_registry, typename X>
            static constexpr decltype(auto) apply(X_registry& p_registry, X&& x)
            {
                return variant(std::make_shared<impl>(
                    p_registry.template get<cppmp::decay_t<T>>(),
                    std::forward<X>(x)));
            }
        };

        /* variant_builder - lvalue reference */

        template<typename T_registry, typename T>
        struct variant_builder<
            cppmp::list<T_registry, T>,
            cppmp::enable_if_t<
                    cppmp::is_same_v<cppmp::decay_t<T_registry>, registry>
                &&  cppmp::is_lvalue_reference_v<T>
            >
        >
        {
            struct impl
                : public variant_impl
            {
                template<typename X>
                inline impl(const cpprtti::type& p_type, X&& x)
                    : variant_impl(p_type)
                {
                    data            = reinterpret_cast<void*>(const_cast<cppmp::decay_t<T>*>(&x));
                    is_const        = cppmp::is_const_v<cppmp::remove_reference_t<T>>;
                    is_reference    = true;
                }

                void assign_impl(const variant& v) override
                {

                };

                template<typename X = cppmp::decay_t<T>>
                cppmp::enable_if_t<cppmp::is_copy_assignable_v<X>>
                assign_helper(const variant& v)
                {
                    *static_cast<cppmp::decay_t<T>*>(data) = v.get<const T&>();
                }

                template<typename X = cppmp::decay_t<T>>
                cppmp::enable_if_t<!cppmp::is_copy_assignable_v<X>>
                assign_helper(const variant& v)
                {
                    throw exception(
                        error_code::variant_assign,
                        cppcore::string_builder()
                            << "Value of type '"
                            << type.name()
                            << "' is not copy assignable!");
                }
            };

            template<typename X_registry, typename X>
            static constexpr decltype(auto) apply(X_registry& p_registry, X&& x)
            {
                return variant(std::make_shared<impl>(
                    p_registry.template get<cppmp::decay_t<T>>(),
                    std::forward<X>(x)));
            }
        };

        /* variant_builder - variant */

        template<typename T_registry, typename T>
        struct variant_builder<
            cppmp::list<T_registry, T>,
            cppmp::enable_if_t<
                    cppmp::is_same_v<cppmp::decay_t<T_registry>, registry>
                &&  cppmp::is_same_v<cppmp::decay_t<T_registry>, variant>
            >
        >
        {
            template<typename X_registry, typename X>
            static constexpr decltype(auto) apply(X_registry& p_registry, X&& x)
                { return std::forward<X>(x); }
        };

        /* variant_builder - type, variant (convert_from) */

        template<typename T_registry, typename... T_args>
        struct variant_builder<
            cppmp::list<T_registry, const type&, T_args...>,
            void
        >
        {
            template<typename X_registry, typename... X_args>
            static constexpr decltype(auto) apply(X_registry& p_registry, const type& t, X_args&&... args)
                { return make_variant(std::forward<X_args>(args)...).to(t.id()); }
        };

        /* value_extractor - default */

        template<typename X, typename = void>
        struct value_extractor
        {
            using is_default = cppmp::true_t;

            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for variant.get(...)!"); }
        };

        /* value_extractor - T&& */

        template<typename T>
        struct value_extractor<
            cppmp::list<
                const variant_impl&,
                const cppmp::type_t<T&&>&>,
            void>
        {
            static constexpr decltype(auto) apply(
                const variant_impl& impl,
                const cppmp::type_t<T&&>&)
            {
                if (impl.is_const)
                {
                    throw exception(
                        error_code::variant_invalid_cast,
                        "Can not move constant value");
                }
                return std::move(*reinterpret_cast<T*>(impl.data));
            }
        };

        /* value_extractor - normal */

        template<typename T>
        struct value_extractor<
            cppmp::list<
                const variant_impl&,
                const cppmp::type_t<T>&>,
            void>
        {
            static constexpr decltype(auto) apply(
                const variant_impl& impl,
                const cppmp::type_t<T>&)
            {
                bool is_const     = cppmp::is_const_v<cppmp::remove_reference_t<T>>;
                bool is_reference = cppmp::is_reference_v<T>;

                if (    is_reference
                    && !is_const
                    &&  impl.is_const)
                {
                    throw exception(
                        error_code::variant_invalid_cast,
                        "Can not get a reference to a const value");
                }

                return *reinterpret_cast<cppmp::remove_reference_t<T>*>(impl.data);
            }
        };

        constexpr decltype(auto) extract_value = cppmp::generic_predicate<__impl::value_extractor> { };

    }

    /* variant */

    variant::variant(__impl::variant_impl_ptr_s p_impl)
        : _impl(p_impl)
        { }

    variant::variant()
        : _impl()
        { }

    const type& variant::type() const
    {
        check_impl();
        return _impl->type;
    }

    bool variant::is_reference() const
    {
        check_impl();
        return _impl->is_reference;
    }

    bool variant::is_const() const
    {
        check_impl();
        return _impl->is_const;
    }

    template<typename T>
    T variant::get() const
    {
        check_impl();

        if (!check_type(_impl->type, get_type_id<cppmp::decay_t<T>>()))
        {
            throw exception(
                error_code::variant_type_mismatch,
                cppcore::string_builder()
                    << "The vaiant does not store the requested type (requested="
                    << get_type_name<cppmp::decay_t<T>>()
                    << ", stored="
                    << _impl->type.name()
                    << ")");
        }

        auto& impl = static_cast<const __impl::variant_impl&>(*_impl);
        return __impl::extract_value(impl, cppmp::type_v<T>);
    }

    bool variant::empty() const
        { return !static_cast<bool>(_impl); }

    template<typename T_to>
    variant variant::to() const
    {
        using to_type = cppmp::decay_t<T_to>;
        auto * t = type().registry().find<to_type>();
        if (!t)
        {
            throw exception(
                error_code::variant_is_empty,
                cppcore::string_builder()
                    << "No known conversion from type '"
                    << type().name()
                    << "' to type '"
                    << get_type_name<to_type>()
                    << "'!");
        }
        return to(*t);
    }

    variant variant::to(cpprtti::type_id id) const
    {
        auto t = type().registry().find(id);
        if (!t)
        {
            throw exception(
                error_code::variant_is_empty,
                cppcore::string_builder()
                    << "No known conversion from type '"
                    << type().name()
                    << "' to type id '"
                    << id
                    << "'!");
        }
        return to(*t);
    }

    variant variant::to(const cpprtti::type& t) const
    {
        check_impl();

        auto& type = _impl->type;
        if (t.id() == type.id())
            return *this;

        auto* converter = type.convert_to(t.id());
        if (!converter)
        {
            throw exception(
                error_code::variant_is_empty,
                cppcore::string_builder()
                    << "No known conversion from type '"
                    << type.name()
                    << "' to type '"
                    << t.name()
                    << "'!");
        }

        return (*converter)(*this);
    }

    inline void variant::assign(const variant& v)
    {
        check_impl();
        _impl->assign(v);
    }

    variant::operator bool() const
        { return static_cast<bool>(_impl); }

    void variant::check_impl() const
    {
        if (!_impl)
        {
            throw exception(
                error_code::variant_is_empty,
                "The variant does not store any value!");
        }
    }

    bool variant::check_type(const cpprtti::type& t, type_id id) const
    {
        if (t.id() == id)
            return true;

        if (t.rtti_type() == cpprtti::rtti_type::class_type)
        {
            auto* _ct = dynamic_cast<const cpprtti::class_type*>(&t);
            assert(_ct);
            auto& ct = *_ct;

            for (auto& kvp : ct.base_classes())
            {
                if (check_type(*kvp, id))
                    return true;
            }
        }

        return false;
    }

}
