#pragma once

#include <map>
#include <memory>

#include "types/type.h"
#include "types/type_decorator.h"

namespace cpprtti
{

    struct registry
    {
    public:
        using type_ptr_s    = std::shared_ptr<type>;
        using type_id_map   = std::map<type_id, type_ptr_s>;
        using type_name_map = std::map<std::string, type_ptr_s>;

    private:
        type_id_map     _id_map;
        type_name_map   _name_map;

    public:
        /**
         * @brief Get a type from the registry.
         *
         * This method will return the type for the requested parameter.
         * It may create a new type object with the default implementation,
         * if the type does not exist.
         */
        template<typename T_type>
        inline auto& get();

        /**
         * @brief Get a type from the registry.
         *
         * This method will return the type for the requested parameter.
         * If the type was not found, an exception is thrown.
         */
        template<typename T_type>
        inline const auto& get() const;

        /**
         * @brief Get a type from the registry.
         *
         * This method will return the type for the requested parameter.
         * It may create a new type object with the passed implementation,
         * if the type does not exist.
         * If the type implementation does not match the requested one,
         * an exception is thrown.
         */
        template<typename T_type, typename T_impl>
        inline auto& get();

        /**
         * @brief Search a type by it's represented C++ type.
         */
        template<typename T_type>
        inline type * find();

        /**
         * @brief Search a type by it's represented C++ type.
         */
        template<typename T_type>
        inline const type * find() const;

        /**
         * @brief Search a type by it's ID.
         */
        inline type * find(type_id id);

        /**
         * @brief Search a type by it's ID.
         */
        inline const type * find(type_id id) const;

        /**
         * @brief Search a type by it's name.
         */
        inline type * find(const std::string& name);

        /**
         * @brief Search a type by it's name.
         */
        inline const type * find(const std::string& name) const;

    private:
        /**
         * @brief Create a new type for the passed C++ type with the passed implementation type.
         */
        template<typename T_impl>
        inline type_decorator<T_impl>& create();
    };

}
