#pragma once

#include <cppmp/misc/getter.h>
#include <cppmp/misc/setter.h>

#include "member_variable.h"

namespace cpprtti
{

    /* member_variable */

    member_variable::member_variable(
            cpprtti::type&      p_owner,
            const std::string&  p_name,
            cpprtti::type&      p_type)
        : member(
            p_owner,
            p_name,
            member_type::variable)
        , _type(p_type)
        { }

    const cpprtti::type& member_variable::type() const
        { return _type; }

    bool member_variable::is_readale() const
        { return _is_readale; }

    bool member_variable::is_writable() const
        { return _is_writable; }

    template<typename T_object>
    variant member_variable::get(T_object& obj) const
    {
        if (!_is_readale)
        {
            throw exception(
                error_code::class_member_not_readable,
                cppcore::string_builder()
                    << "Member variable '"
                    << _name
                    << "' of '"
                    << _owner.name()
                    << " is not readable!");
        }
        return get_impl(make_variant(_owner.registry(), obj));
    }

    variant member_variable::get(const variant& obj) const
        { return get_impl(obj); }

    template<typename T_object, typename T_value>
    inline void member_variable::set(T_object& obj, const T_value& value) const
    {
        if (!_is_writable)
        {
            throw exception(
                error_code::class_member_not_writable,
                cppcore::string_builder()
                    << "Member variable '"
                    << _name
                    << "' of '"
                    << _owner.name()
                    << " is not writable!");
        }
        return set_impl(make_variant(_owner.registry(), obj), make_variant(_owner.registry(), value));
    }

    void member_variable::set(const variant& obj, const variant& value) const
        { return set_impl(obj, value); }

    namespace __impl
    {

        /* getter_info_trait */

        template<typename T_getter, typename = void>
        struct getter_info_trait;

        template<typename T_getter>
        struct getter_info_trait<
            T_getter,
            cppmp::enable_if_t<
                decltype(cppmp::make_getter)::is_valid_v<T_getter>
            >
        >
        {
            using getter_type = decltype(cppmp::make_getter(std::declval<T_getter>()));
            using object_type = typename getter_type::object_type;
            using value_type  = typename getter_type::value_type;

            static constexpr decltype(auto) is_read_only_v =
                    cppmp::is_const_v<cppmp::remove_reference_t<object_type>>
                ||  cppmp::is_const_v<cppmp::remove_reference_t<value_type>>
                || !cppmp::is_copy_assignable_v<value_type>;
        };

        /* setter_info_trait */

        template<typename T_setter, typename = void>
        struct setter_info_trait;

        template<typename T_setter>
        struct setter_info_trait<
            T_setter,
            cppmp::enable_if_t<
                decltype(cppmp::make_setter)::is_valid_v<T_setter>
            >
        >
        {
            using setter_type = decltype(cppmp::make_setter(std::declval<T_setter>()));
            using object_type = typename setter_type::object_type;
            using value_type  = typename setter_type::value_type;

            static constexpr decltype(auto) is_write_only_v =
                !cppmp::is_base_of_v<cppmp::tag_setter_member_var, setter_type>;
        };

        /* member_variable_builder - default */

        template<typename X, typename>
        struct member_variable_builder
        {
            using is_default = cppmp::true_t;

            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for make_member_variable(...)!"); }
        };

        /* member_variable_builder - read-write getter */

        template<typename T_getter>
        struct member_variable_builder<
            cppmp::list<class_type&, const std::string&, T_getter>,
            cppmp::enable_if_t<
                    decltype(cppmp::make_getter)::is_valid_v<T_getter>
                && !getter_info_trait<T_getter>::is_read_only_v
            >
        >
        {
            using getter_type = cppmp::decay_t<decltype(cppmp::make_getter(std::declval<T_getter>()))>;
            using object_type = typename getter_type::object_type;
            using value_type  = typename getter_type::value_type;

            struct member_impl
                : public member_variable
            {
            private:
                getter_type _getter;

            public:
                inline member_impl(
                    cpprtti::type&      p_owner,
                    const std::string&  p_name,
                    getter_type         p_getter)
                    : member_variable(p_owner, p_name, p_owner.registry().get<value_type>())
                    , _getter(p_getter)
                {
                    if (_owner.id() != get_type_id<cppmp::decay_t<object_type>>())
                    {
                        throw exception(
                            error_code::class_type_mismatch,
                            "Object type of the getter does not match the type of the owner!");
                    }

                    _is_readale  = true;
                    _is_writable = true;
                }

            protected:
                variant get_impl(const variant& obj) const override
                {
                    return obj.is_const()
                        ? make_variant(_owner.registry(), _getter(obj.get<const object_type&>()))
                        : make_variant(_owner.registry(), _getter(obj.get<object_type&>()));
                }

                void set_impl(const variant& obj, const variant& value) const override
                    { _getter(obj.get<object_type&>()) = value.get<value_type&>(); }
            };

            template<typename X_getter>
            static constexpr decltype(auto) apply(type& owner, const std::string& name, X_getter&& getter)
            {
                return std::make_unique<member_impl>(
                    owner,
                    name,
                    cppmp::make_getter(std::forward<X_getter>(getter)));
            }
        };

        /* member_variable_builder - read-only getter */

        template<typename T_getter>
        struct member_variable_builder<
            cppmp::list<class_type&, const std::string&, T_getter>,
            cppmp::enable_if_t<
                    decltype(cppmp::make_getter)::is_valid_v<T_getter>
                &&  getter_info_trait<T_getter>::is_read_only_v
            >
        >
        {
            using getter_type = cppmp::decay_t<decltype(cppmp::make_getter(std::declval<T_getter>()))>;
            using object_type = typename getter_type::object_type;
            using value_type  = typename getter_type::value_type;

            struct member_impl
                : public member_variable
            {
            private:
                getter_type _getter;

            public:
                inline member_impl(
                    cpprtti::type&      p_owner,
                    const std::string&  p_name,
                    getter_type         p_getter)
                    : member_variable(p_owner, p_name, p_owner.registry().get<value_type>())
                    , _getter(p_getter)
                {
                    if (_owner.id() != get_type_id<cppmp::decay_t<object_type>>())
                    {
                        throw exception(
                            error_code::class_type_mismatch,
                            "Object type of the getter does not match the type of the owner!");
                    }

                    _is_readale  = true;
                    _is_writable = false;
                }

            protected:
                variant get_impl(const variant& obj) const override
                    { return make_variant(_owner.registry(), _getter(obj.get<const object_type&>())); }

                void set_impl(const variant& obj, const variant& value) const override
                    { }
            };

            template<typename X_getter>
            static constexpr decltype(auto) apply(type& owner, const std::string& name, X_getter&& getter)
            {
                return std::make_unique<member_impl>(
                    owner,
                    name,
                    cppmp::make_getter(std::forward<X_getter>(getter)));
            }
        };

        /* member_variable_builder - write-only setter */

        template<typename T_setter>
        struct member_variable_builder<
            cppmp::list<class_type&, const std::string&, T_setter>,
            cppmp::enable_if_t<
                    decltype(cppmp::make_setter)::is_valid_v<T_setter>
                &&  setter_info_trait<T_setter>::is_write_only_v
            >
        >
        {
            using setter_type = cppmp::decay_t<decltype(cppmp::make_setter(std::declval<T_setter>()))>;
            using object_type = typename setter_type::object_type;
            using value_type  = typename setter_type::value_type;

            struct member_impl
                : public member_variable
            {
            private:
                setter_type _setter;

            public:
                inline member_impl(
                    cpprtti::type&      p_owner,
                    const std::string&  p_name,
                    setter_type         p_setter)
                    : member_variable(p_owner, p_name, p_owner.registry().get<value_type>())
                    , _setter(p_setter)
                {
                    if (_owner.id() != get_type_id<cppmp::decay_t<object_type>>())
                    {
                        throw exception(
                            error_code::class_type_mismatch,
                            "Object type of the setter does not match the type of the owner!");
                    }

                    _is_readale  = false;
                    _is_writable = true;
                }

            protected:
                variant get_impl(const variant& obj) const override
                    { return variant(); }

                void set_impl(const variant& obj, const variant& value) const override
                    { _setter(obj.get<object_type&>(), value.get<const value_type&>()); }
            };

            template<typename X_setter>
            static constexpr decltype(auto) apply(type& owner, const std::string& name, X_setter&& setter)
            {
                return std::make_unique<member_impl>(
                    owner,
                    name,
                    cppmp::make_setter(std::forward<X_setter>(setter)));
            }
        };

        /* member_variable_builder - read-write getter/setter */

        template<typename T_getter, typename T_setter>
        struct member_variable_builder<
            cppmp::list<class_type&, const std::string&, T_getter, T_setter>,
            cppmp::enable_if_t<
                    decltype(cppmp::make_getter)::is_valid_v<T_getter>
                &&  decltype(cppmp::make_setter)::is_valid_v<T_setter>
            >
        >
        {
            using getter_type = cppmp::decay_t<decltype(cppmp::make_getter(std::declval<T_getter>()))>;
            using getter_object_type = typename getter_type::object_type;
            using getter_value_type  = typename getter_type::value_type;

            using setter_type = cppmp::decay_t<decltype(cppmp::make_setter(std::declval<T_setter>()))>;
            using setter_object_type = typename setter_type::object_type;
            using setter_value_type  = typename setter_type::value_type;

            using object_type = cppmp::decay_t<getter_object_type>;
            using value_type  = cppmp::decay_t<getter_value_type>;

            static_assert(cppmp::is_same_v<
                cppmp::decay_t<getter_object_type>,
                cppmp::decay_t<setter_object_type>>,
                "object type does not match");
            static_assert(cppmp::is_same_v<
                cppmp::decay_t<getter_value_type>,
                cppmp::decay_t<setter_value_type>>,
                "value type does not match");

            struct member_impl
                : public member_variable
            {
            private:
                getter_type _getter;
                setter_type _setter;

            public:
                inline member_impl(
                    cpprtti::type&      p_owner,
                    const std::string&  p_name,
                    getter_type         p_getter,
                    setter_type         p_setter)
                    : member_variable(p_owner, p_name, p_owner.registry().get<value_type>())
                    , _getter(p_getter)
                    , _setter(p_setter)
                {
                    if (_owner.id() != get_type_id<cppmp::decay_t<object_type>>())
                    {
                        throw exception(
                            error_code::class_type_mismatch,
                            "Object type of the setter does not match the type of the owner!");
                    }

                    _is_readale  = true;
                    _is_writable = true;
                }

            protected:
                variant get_impl(const variant& obj) const override
                {
                    return obj.is_const()
                        ? make_variant(_owner.registry(), _getter(obj.get<const getter_object_type&>()))
                        : make_variant(_owner.registry(), _getter(obj.get<getter_object_type&>()));
                }

                void set_impl(const variant& obj, const variant& value) const override
                    { _setter(obj.get<setter_object_type&>(), value.get<setter_value_type>()); }
            };

            template<typename X_getter, typename X_setter>
            static constexpr decltype(auto) apply(type& owner, const std::string& name, X_getter&& getter, X_setter&& setter)
            {
                return std::make_unique<member_impl>(
                    owner,
                    name,
                    cppmp::make_getter(std::forward<X_getter>(getter)),
                    cppmp::make_setter(std::forward<X_setter>(setter)));
            }
        };

    }

}
