#pragma once

#include <map>
#include <string>
#include <functional>

#include <cpprtti/misc/helper.h>
#include <cpprtti/misc/variant.h>
#include <cpprtti/misc/storage.h>

namespace cpprtti
{

    struct registry;

    enum class rtti_type
    {
        unknown = 0,        //!< unknown
        fundamental_type,   //!< fundamental C++ type (like int, float, ...)
        class_type,         //!< C++ class/struct
    };

    using rtti_type_t = enum rtti_type;

    /**
     * @brief represents any C++ type
     */
    struct type
        : public storage_container
    {
    public:
        using converter     = std::function<variant(variant)>;
        using converter_map = std::map<type_id, converter>;

        template<typename T_base>
        friend struct type_decorator;

    protected:
        cpprtti::registry&  _registry;      //!< Type registry this type belongs to.
        type_id             _id;            //!< Unique ID
        std::string         _name;          //!< Name of the type.
        rtti_type_t         _rtti_type;     //!< Type specialization.
        converter_map       _converters;    //!< Convert this type to another type.

    public:
        /**
         * @brief Constructor.
         */
        inline type(
            cpprtti::registry&  p_registry,
            size_t              p_id,
            const std::string&  p_name,
            rtti_type_t         p_rtti_type);

        /**
         * @brief Destructor.
         */
        virtual ~type() = default;

        /**
         * @brief Type registry this type belongs to.
         */
        inline cpprtti::registry& registry();

        /**
         * @brief Type registry this type belongs to.
         */
        inline const cpprtti::registry& registry() const;

        /**
         * @brief Get the ID of this type.
         */
        inline type_id id() const;

        /**
         * @brief Get the name of this type.
         */
        inline const std::string& name() const;

        /**
         * @brief Get the RTTI type of this type object.
         */
        inline rtti_type_t rtti_type() const;

        /**
         * @brief Get a converter that can convert this type to the passed type.
         */
        template<typename T_to>
        inline const converter* convert_to() const;

        /**
         * @brief Get a converter that can convert this type to the passed type ID.
         */
        inline const converter* convert_to(type_id id) const;

        /**
         * @brief Get a converter that can convert the passed type to this type.
         */
        template<typename T_from>
        inline const converter* convert_from() const;

        /**
         * @brief Get a converter that can convert the passed type ID to this type.
         */
        inline const converter* convert_from(type_id id) const;
    };

}
